# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``init`` command."""
import os
from pathlib import Path

import pytest

from renku.cli import cli
from renku.cli.init import create_template_sentence, parse_parameters
from renku.core import errors
from tests.utils import raises


def test_parse_parameters(project_init):
    def clean_param(p):
        return [v for v in p if v != "--parameter"]

    data, commands = project_init

    parsed = parse_parameters(None, None, clean_param(commands["parameters"]))
    keys = parsed.keys()
    assert 2 == len(keys)
    assert "p1" in keys
    assert "p2" in keys
    assert "v1" == parsed["p1"]
    assert "v2" == parsed["p2"]
    with raises(errors.ParameterError):
        parse_parameters(
            None, None, clean_param(commands["parameters"]) + clean_param(commands["parameters_equal_missing"])
        )
    with raises(errors.ParameterError):
        parse_parameters(
            None, None, clean_param(commands["parameters"]) + clean_param(commands["parameters_equal_early"])
        )


def test_template_selection_helpers():
    templates = [
        {"name": "Template Python", "folder": "folder_python", "description": "Description Python"},
        {
            "name": "Template R",
            "folder": "folder_R",
            "description": "Description R",
            "variables": {"custom": "random data"},
        },
    ]
    instructions = "Please choose a template by typing the index"
    sentence = create_template_sentence(templates)
    stripped_sentence = " ".join(sentence.split())

    assert "1 folder_python" in stripped_sentence
    assert "2 folder_R" in stripped_sentence
    assert instructions not in stripped_sentence
    full_sentence = create_template_sentence(templates, instructions=True)
    assert instructions in full_sentence

    # with describe=True
    sentence = create_template_sentence(templates, describe=True)
    stripped_sentence = " ".join(sentence.split())

    assert "1 folder_python Description Python" in stripped_sentence
    assert "2 folder_R custom: random data Description R" in stripped_sentence
    assert instructions not in stripped_sentence

    full_sentence = create_template_sentence(templates, describe=True, instructions=True)
    assert instructions in full_sentence


def test_list_templates(isolated_runner, project_init, template):
    """Test listing templates."""
    data, commands = project_init

    new_project = Path(data["test_project"])
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, commands["init_test"] + commands["list"])
    assert 0 == result.exit_code
    assert not new_project.exists()
    assert template["id"] in result.output


def test_init(isolated_runner, project_init):
    """Test project initialization from template."""
    data, commands = project_init

    # create the project
    new_project = Path(data["test_project"])
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"], commands["confirm"])
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / ".renku").exists()
    assert (new_project / ".renku" / "renku.ini").exists()
    assert (new_project / ".renku" / "metadata.yml").exists()

    # try to re-create in the same folder
    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"], commands["confirm"])
    assert 0 != result.exit_code

    # force re-create in the same folder
    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["force"], commands["confirm"]
    )
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / ".renku").exists()
    assert (new_project / ".renku" / "renku.ini").exists()
    assert (new_project / ".renku" / "metadata.yml").exists()

    # init using index instead of id
    new_project_2 = Path(data["test_project_alt"])
    result = isolated_runner.invoke(cli, commands["init_alt"] + commands["index"], commands["confirm"])
    assert 0 == result.exit_code
    assert new_project_2.exists()
    assert (new_project_2 / ".renku").exists()
    assert (new_project_2 / ".renku" / "renku.ini").exists()
    assert (new_project_2 / ".renku" / "metadata.yml").exists()

    # verify both init lead to the same result
    template_files = [f for f in new_project.glob("**/*") if ".git" not in str(f)]
    for template_file in template_files:
        expected_file = new_project_2 / template_file.relative_to(new_project)
        assert expected_file.exists()

    # verify providing both index and id fails
    result = isolated_runner.invoke(
        cli, commands["init_alt"] + commands["index"] + commands["id"] + commands["force"], commands["confirm"]
    )
    assert 2 == result.exit_code
    assert "Use either --template-id or --template-index, not both" in result.output


def test_init_force_in_empty_dir(isolated_runner, project_init):
    """Run init --force in empty directory."""
    data, commands = project_init

    new_project = Path(data["test_project"])
    assert not new_project.exists()
    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["force"], commands["confirm"]
    )
    assert 0 == result.exit_code


def test_init_force_in_dirty_dir(isolated_runner, project_init):
    """Run init --force in dirty directory."""
    data, commands = project_init

    new_project = Path(data["test_project"])
    assert not new_project.exists()

    new_project.mkdir(parents=True)
    random_file = new_project / "random_file.txt"
    with random_file.open("w") as dest:
        dest.writelines(["random text"])
    assert random_file.exists()

    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"], commands["confirm"])
    lines = result.output.split("\n")
    assert 1 == result.exit_code
    assert "is not empty" in lines[0]
    assert "random_file.txt" in lines[1]

    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["force"], commands["confirm"]
    )
    assert random_file.exists()
    assert 0 == result.exit_code


def test_init_on_cloned_repo(isolated_runner, data_repository, project_init):
    """Run init --force in directory containing another repo."""
    data, commands = project_init

    new_project = Path(data["test_project"])
    import shutil

    shutil.copytree(str(data_repository.working_dir), str(new_project))
    assert new_project.exists()

    # try to create in a dirty folder
    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"], commands["confirm"])
    assert 0 != result.exit_code

    # force re-create in the same folder
    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["force"], commands["confirm"]
    )
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / ".renku").exists()
    assert (new_project / ".renku" / "renku.ini").exists()
    assert (new_project / ".renku" / "metadata.yml").exists()


@pytest.mark.integration
def test_init_remote(isolated_runner, project_init):
    """Test project initialization from a remote template."""
    data, commands = project_init

    # create the project
    new_project = Path(data["test_project"])
    assert not new_project.exists()
    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["force"], commands["confirm"]
    )
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / ".renku").exists()
    assert (new_project / ".renku" / "renku.ini").exists()
    assert (new_project / ".renku" / "metadata.yml").exists()


def test_init_with_parameters(isolated_runner, project_init, template):
    """Test project initialization using custom metadata."""
    data, commands = project_init

    # create the project
    new_project = Path(data["test_project"])
    assert not new_project.exists()
    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["parameters"] + commands["parameters_equal_missing"]
    )
    assert 0 != result.exit_code
    assert (
        f'Error: Invalid parameter value for --parameter "{ commands["parameters_equal_missing"][1]}"' in result.output
    )

    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["parameters"] + commands["parameters_equal_early"]
    )
    assert 0 != result.exit_code
    assert f'Error: Invalid parameter value for --parameter "{commands["parameters_equal_early"][1]}"' in result.output

    result = isolated_runner.invoke(
        cli, commands["init_test"] + commands["id"] + commands["parameters"], commands["confirm"]
    )
    assert 0 == result.exit_code
    assert "The template requires a value for" in result.output
    for param in set(template["metadata"].keys()):
        assert param in result.output
    assert "These parameters are not used by the template and were ignored:" in result.output


@pytest.mark.parametrize("data_dir", ["dir", "nested/dir/s"])
def test_init_with_data_dir(isolated_runner, data_dir, directory_tree, project_init):
    """Test initializing with data directory."""
    from git import Repo

    data, commands = project_init

    new_project = Path(data["test_project"])
    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"] + ["--data-dir", data_dir])
    assert 0 == result.exit_code

    assert (new_project / data_dir).exists()
    assert (new_project / data_dir / ".gitkeep").exists()
    assert not Repo(new_project).is_dirty()

    os.chdir(new_project.resolve())
    result = isolated_runner.invoke(cli, ["dataset", "add", "-c", "my-data", str(directory_tree)])
    assert 0 == result.exit_code, result.output
    assert (Path(data_dir) / "my-data" / directory_tree.name / "file1").exists()


@pytest.mark.parametrize("data_dir", ["/absolute/path/outside", "../relative/path/outside"])
def test_init_with_wrong_data_dir(isolated_runner, data_dir, project_init):
    """Test initialization fails with wrong data directory."""
    data, commands = project_init

    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"] + ["--data-dir", data_dir])
    assert 2 == result.exit_code
    assert f"Data directory {data_dir} is not within project" in result.output


@pytest.mark.parametrize("data_dir", [".", ".git", ".renku", ".git/"])
def test_init_with_invalid_data_dir(isolated_runner, data_dir, project_init):
    """Test initialization fails with invalid data directory."""
    data, commands = project_init

    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"] + ["--data-dir", data_dir])
    assert 2 == result.exit_code
    data_dir = data_dir.rstrip("/")
    assert f"Cannot use {data_dir} as data directory." in result.output


def test_default_init_parameters(isolated_runner, mocker, project_init, template):
    """Test that the default parameters are set in template initialisation."""
    create_from_template = mocker.patch("renku.cli.init.create_from_template")
    mocker.patch("renku.cli.githooks.install")

    data, commands = project_init

    new_project = Path(data["test_project"])
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, commands["init_test"] + commands["id"], commands["confirm"])
    assert 0 == result.exit_code
    create_from_template.assert_called_once()
    metadata = create_from_template.call_args[1]["metadata"]
    assert {
        "__template_source__",
        "__template_ref__",
        "__template_id__",
        "__namespace__",
        "__repository__",
        "__project_slug__",
        "__sanitized_project_name__",
    } <= set(metadata.keys())
    assert metadata["__template_source__"] == "renku"
    assert metadata["__template_ref__"] == "master"
    assert metadata["__template_id__"] == template["id"]
    assert metadata["__namespace__"] == ""
    assert metadata["__repository__"] == ""
    assert metadata["__project_slug__"] == ""
    assert metadata["__sanitized_project_name__"] == ""



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git regex tests."""

import pytest

from renku.core import errors
from renku.core.models.git import GitURL


@pytest.mark.parametrize(
    "fields",
    [
        {
            "href": "https://example.com/repo.git",
            "protocol": "https",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "repo.git",
        },
        {
            "href": "https://example.com/repo",
            "protocol": "https",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "repo",
        },
        {
            "href": "https://example.com/owner/repo.git",
            "protocol": "https",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "owner/repo.git",
            "owner": "owner",
        },
        {
            "href": "https://example.com:1234/repo.git",
            "protocol": "https",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "repo.git",
            "port": "1234",
        },
        {
            "href": "https://example.com:1234/owner/repo.git",
            "protocol": "https",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "owner/repo.git",
            "owner": "owner",
            "port": "1234",
        },
        {
            "href": "https://example.com:1234/prefix/owner/repo.git",
            "protocol": "https",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "prefix/owner/repo.git",
            "owner": "owner",
            "port": "1234",
        },
        {
            "href": "https://example.com/pre.fix/owner.name/repo.name.git",
            "protocol": "https",
            "hostname": "example.com",
            "pathname": "pre.fix/owner.name/repo.name.git",
            "owner": "owner.name",
            "name": "repo.name",
        },
        {
            "href": "git+https://example.com:1234/owner/repo.git",
            "protocol": "git+https",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "owner/repo.git",
            "owner": "owner",
            "port": "1234",
        },
        {
            "href": "git+ssh://example.com:1234/owner/repo.git",
            "protocol": "git+ssh",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "owner/repo.git",
            "owner": "owner",
            "port": "1234",
        },
        {
            "href": "git+ssh://user:pass@example.com:1234/owner/repo.git",
            "protocol": "git+ssh",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "owner/repo.git",
            "owner": "owner",
            "port": "1234",
            "username": "user",
            "password": "pass",
        },
        {
            "href": "ssh://user:pass@example.com/~user/owner/repo.git",
            "protocol": "ssh",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "~user/owner/repo.git",
            "owner": "owner",
            "username": "user",
            "password": "pass",
        },
        pytest.param(
            {
                "href": "git@example.com/repo.git",
                "protocol": "ssh",
                "hostname": "example.com",
                "name": "repo",
                "pathname": "repo.git",
                "username": "git",
            },
            marks=pytest.mark.xfail(raises=errors.ConfigurationError, strict=True),
        ),
        pytest.param(
            {
                "href": "git@example.com/owner/repo.git",
                "protocol": "ssh",
                "hostname": "example.com",
                "name": "repo",
                "pathname": "owner/repo.git",
                "owner": "owner",
                "username": "git",
            },
            marks=pytest.mark.xfail(raises=errors.ConfigurationError, strict=True),
        ),
        {
            "href": "git@example.com:repo.git",
            "protocol": "ssh",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "repo.git",
            "username": "git",
        },
        {
            "href": "git@example.com:owner/repo.git",
            "protocol": "ssh",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "owner/repo.git",
            "owner": "owner",
            "username": "git",
        },
        {
            "href": "git@example.com:prefix/owner/repo.git",
            "protocol": "ssh",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "prefix/owner/repo.git",
            "owner": "owner",
            "username": "git",
        },
        {"href": "/path/to/repo", "pathname": "/path/to/repo",},
        {"href": "file:///path/to/repo", "pathname": "/path/to/repo",},
        {"href": "../relative/path/to/repo", "pathname": "../relative/path/to/repo",},
        {"href": "file://../relative/path/to/repo", "pathname": "../relative/path/to/repo",},
        pytest.param(
            {
                "href": "https://example.com:1234:repo.git",
                "protocol": "https",
                "hostname": "example.com",
                "port": "1234",
                "name": "repo",
                "pathname": "repo.git",
            },
            marks=pytest.mark.xfail(raises=errors.ConfigurationError, strict=True),
        ),
        pytest.param(
            {
                "href": "https://example.com:1234:owner/repo.git",
                "protocol": "https",
                "hostname": "example.com",
                "port": "1234",
                "name": "repo",
                "pathname": "repo.git",
                "owner": "owner",
            },
            marks=pytest.mark.xfail(raises=errors.ConfigurationError, strict=True),
        ),
        pytest.param(
            {
                "href": "git@example.com:1234:owner/repo.git",
                "protocol": "ssh",
                "hostname": "example.com",
                "port": "1234",
                "name": "repo",
                "pathname": "repo.git",
                "owner": "owner",
            },
            marks=pytest.mark.xfail(raises=errors.ConfigurationError, strict=True),
        ),
        {
            "href": "git@example.com:1234/prefix/owner/repo.git",
            "username": "git",
            "protocol": "ssh",
            "hostname": "example.com",
            "name": "repo",
            "pathname": "1234/prefix/owner/repo.git",
            "owner": "owner",
        },
    ],
)
def test_valid_href(fields):
    """Test the various repo regexes."""
    fields.pop("protocols", None)
    assert GitURL(**fields) == GitURL.parse(fields["href"])



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``rerun`` command."""

from __future__ import absolute_import, print_function

import os
import subprocess
from pathlib import Path

import git
import pytest
from click.testing import CliRunner

from renku.cli import cli


def test_simple_rerun(runner, project, run, no_lfs_warning):
    """Test simple file recreation."""
    greetings = {"hello", "hola", "ahoj"}

    cwd = Path(project)
    source = cwd / "source.txt"
    selected = cwd / "selected.txt"

    repo = git.Repo(project)

    with source.open("w") as f:
        f.write("\n".join(greetings))

    repo.git.add("--all")
    repo.index.commit("Created greetings")

    cmd = ["run", "python", "-S", "-c", "import sys, random; print(random.choice(sys.stdin.readlines()))"]

    assert 0 == run(cmd, stdin=source, stdout=selected)

    with selected.open("r") as f:
        greeting = f.read().strip()
        assert greeting in greetings

    def _rerun():
        """Return greeting after reruning."""
        result = run(args=("rerun", str(selected)))
        assert 0 == result
        with selected.open("r") as fp:
            greeting = fp.read().strip()
            assert greeting in greetings
            return greeting

    for _ in range(100):
        new_greeting = _rerun()
        if greeting != new_greeting:
            break

    assert greeting != new_greeting, "Something is not random"

    for _ in range(100):
        new_greeting = _rerun()
        if greeting == new_greeting:
            break

    assert greeting == new_greeting, "Something is not random"


def test_rerun_with_inputs(runner, project, run):
    """Test file recreation with specified inputs."""
    cwd = Path(project)
    first = cwd / "first.txt"
    second = cwd / "second.txt"
    inputs = (first, second)

    output = cwd / "output.txt"

    cmd = ["run", "python", "-S", "-c", "import random; print(random.random())"]

    for file_ in inputs:
        assert 0 == run(args=cmd, stdout=file_), "Random number generation."

    cmd = ["run", "cat"] + [str(path) for path in inputs]
    assert 0 == run(args=cmd, stdout=output)

    with output.open("r") as f:
        initial_data = f.read()

    assert 0 == run(args=("rerun", str(output)))

    with output.open("r") as f:
        assert f.read() != initial_data, "The output should have changed."


def test_rerun_with_inputs_with_from(runner, project, run):
    """Test file recreation with specified inputs."""
    cwd = Path(project)
    first = cwd / "first.txt"
    second = cwd / "second.txt"
    inputs = (first, second)

    output = cwd / "output.txt"

    cmd = ["run", "python", "-S", "-c", "import random; print(random.random())"]

    for file_ in inputs:
        assert 0 == run(args=cmd, stdout=file_), "Random number generation."

    cmd = ["run", "cat"] + [str(path) for path in inputs]
    assert 0 == run(args=cmd, stdout=output)

    # Keep the first file unchanged.
    with first.open("r") as f:
        first_data = f.read()

    assert 0 == run(args=("rerun", "--from", str(first), str(output)))

    with output.open("r") as f:
        assert f.read().startswith(first_data)


def test_rerun_with_edited_inputs(project, run, no_lfs_warning):
    """Test input modification."""
    runner = CliRunner(mix_stderr=False)

    cwd = Path(project)
    data = cwd / "examples"
    data.mkdir()
    first = data / "first.txt"
    second = data / "second.txt"
    third = data / "third.txt"

    run(args=["run", "echo", "hello"], stdout=first)
    run(args=["run", "cat", str(first)], stdout=second)
    run(args=["run", "echo", "1"], stdout=third)

    with first.open("r") as first_fp:
        with second.open("r") as second_fp:
            assert first_fp.read() == second_fp.read()

    # Change the initial input from "hello" to "hola".
    from click.testing import make_input_stream

    stdin = make_input_stream("hola\n", "utf-8")
    assert 0 == run(args=("rerun", "--edit-inputs", str(second)), stdin=stdin)

    with second.open("r") as second_fp:
        assert "hola\n" == second_fp.read()

    # Change the input from examples/first.txt to examples/third.txt.
    stdin = make_input_stream(str(third.name), "utf-8")
    old_dir = os.getcwd()
    try:
        # Make sure the input path is relative to the current directory.
        os.chdir(str(data))

        result = runner.invoke(
            cli, ["rerun", "--show-inputs", "--from", str(first), str(second)], catch_exceptions=False
        )
        assert 0 == result.exit_code
        assert result.output.startswith("https://")
        assert result.output[:-1].endswith(first.name)
        assert 0 == run(args=("rerun", "--edit-inputs", "--from", str(first), str(second)), stdin=stdin)
    finally:
        os.chdir(old_dir)

    with third.open("r") as third_fp:
        with second.open("r") as second_fp:
            assert third_fp.read() == second_fp.read()


@pytest.mark.parametrize("cmd, exit_code", (("update", 0), ("rerun", 1)))
def test_input_update_and_rerun(cmd, exit_code, runner, project, run):
    """Test update and rerun of an input."""
    repo = git.Repo(project)
    cwd = Path(project)
    input_ = cwd / "input.txt"
    with input_.open("w") as f:
        f.write("first")

    repo.git.add("--all")
    repo.index.commit("Created input.txt")

    assert exit_code == run(args=(cmd, input_.name))


def test_output_directory(runner, project, run, no_lfs_size_limit):
    """Test detection of output directory."""
    cwd = Path(project)
    data = cwd / "source" / "data.txt"
    source = data.parent
    source.mkdir(parents=True)
    data.write_text("data")

    # Empty destination
    destination = cwd / "destination"
    source_wc = cwd / "destination_source.wc"
    # Non empty destination
    invalid_destination = cwd / "invalid_destination"
    invalid_destination.mkdir(parents=True)
    (invalid_destination / "non_empty").touch()

    repo = git.Repo(project)
    repo.git.add("--all")
    repo.index.commit("Created source directory", skip_hooks=True)

    cmd = ["run", "cp", "-LRf", str(source), str(destination)]
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert 0 == result.exit_code

    destination_source = destination / data.name
    assert destination_source.exists()

    # check that the output in subdir is added to LFS
    with (cwd / ".gitattributes").open() as f:
        gitattr = f.read()
    assert str(destination.relative_to(cwd)) + "/**" in gitattr
    assert destination_source.name in subprocess.check_output(["git", "lfs", "ls-files"]).decode()

    cmd = ["run", "wc"]
    assert 0 == run(args=cmd, stdin=destination_source, stdout=source_wc)

    # Make sure the output directory can be recreated
    assert 0 == run(args=("rerun", str(source_wc)))
    assert {data.name} == {path.name for path in destination.iterdir()}

    cmd = ["log", str(source_wc)]
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    destination_data = str(Path("destination") / "data.txt")
    assert destination_data in result.output, cmd
    assert " directory)" in result.output

    cmd = ["run", "cp", "-r", str(source), str(invalid_destination)]
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert 1 == result.exit_code
    assert not (invalid_destination / data.name).exists()



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checks needed to determine integrity of datasets."""
import os
from collections import defaultdict

import click

from renku.core.utils.migrate import get_pre_0_3_4_datasets_metadata

from ..echo import WARNING


def check_dataset_metadata(client):
    """Check location of dataset metadata."""
    old_metadata = get_pre_0_3_4_datasets_metadata(client)

    if not old_metadata:
        return True, None

    problems = (
        WARNING + "There are metadata files in the old location."
        '\n  (use "renku migrate" to move them)\n\n\t'
        + "\n\t".join(click.style(str(path.relative_to(client.path)), fg="yellow") for path in old_metadata)
        + "\n"
    )

    return False, problems


def check_missing_files(client):
    """Find missing files listed in datasets."""
    missing = defaultdict(list)

    for dataset in client.datasets.values():
        for file_ in dataset.files:
            path = client.path / file_.path
            file_exists = path.exists() or (file_.external and os.path.lexists(path))
            if not file_exists:
                missing[dataset.name].append(file_.path)

    if not missing:
        return True, None

    problems = WARNING + "There are missing files in datasets."

    for dataset, files in missing.items():
        problems += (
            "\n\t"
            + click.style(dataset, fg="yellow")
            + ":\n\t  "
            + "\n\t  ".join(click.style(path, fg="red") for path in files)
        )

    return False, problems



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``remove`` command."""

from renku.cli import cli
from renku.core.management.repository import DEFAULT_DATA_DIR as DATA_DIR


def test_remove_dataset_file(isolated_runner, client, tmpdir, subdirectory):
    """Test remove of a file that belongs to a dataset."""
    runner = isolated_runner

    # create a dataset
    result = runner.invoke(cli, ["dataset", "create", "testing"])
    assert 0 == result.exit_code
    assert "OK" in result.output

    source = tmpdir.join("remove_dataset.file")
    source.write(DATA_DIR)

    result = runner.invoke(cli, ["dataset", "add", "testing", source.strpath])
    assert 0 == result.exit_code

    path = client.path / client.data_dir / "testing" / "remove_dataset.file"
    assert path.exists()

    result = runner.invoke(cli, ["doctor"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["rm", str(client.path / DATA_DIR)])
    assert 0 == result.exit_code, result.output

    assert not path.exists()

    result = runner.invoke(cli, ["doctor"])
    assert 0 == result.exit_code



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Dataset API."""

import os

import pytest

from renku.api import Dataset, Project


def test_list_datasets(client_with_datasets):
    """Test listing datasets within a project context."""
    with Project():
        datasets = Dataset.list()

        assert {"dataset-1", "dataset-2"} == {d.name for d in datasets}


def test_list_datasets_outside_a_context(client_with_datasets):
    """Test listing datasets outside a project context."""
    datasets = Dataset.list()

    assert {"dataset-1", "dataset-2"} == {d.name for d in datasets}


def test_list_datasets_outside_a_renku_project(directory_tree):
    """Test listing datasets in a non-renku directory."""
    os.chdir(directory_tree)

    assert [] == Dataset.list()


@pytest.mark.parametrize(
    "dataset, files_paths",
    [
        ("dataset-1", []),
        ("dataset-2", ["data/dataset-2/file1", "data/dataset-2/dir1/file2", "data/dataset-2/dir1/file3"]),
    ],
)
def test_list_dataset_files(client_with_datasets, dataset, files_paths):
    """Test listing datasets files."""
    with Project() as project:
        dataset = next(d for d in Dataset.list() if d.name == dataset)

        assert set(files_paths) == {d.path for d in dataset.files}
        assert set(project.path / p for p in files_paths) == {d.full_path for d in dataset.files}



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client utilities."""

import functools
import uuid

import yaml

from renku.core.incubation.command import Command, CommandResult

from .git import get_git_isolation


def _uuid_representer(dumper, data):
    """Add UUID serializer for YAML."""
    return dumper.represent_str(str(data))


yaml.add_representer(uuid.UUID, _uuid_representer)


def pass_local_client(
    method=None,
    clean=None,
    requires_migration=False,
    commit=None,
    commit_only=None,
    ignore_std_streams=True,
    commit_empty=True,
    raise_if_empty=False,
    lock=None,
):
    """Pass client from the current context to the decorated command."""
    if method is None:
        return functools.partial(
            pass_local_client,
            clean=clean,
            requires_migration=requires_migration,
            commit=commit,
            commit_only=commit_only,
            ignore_std_streams=ignore_std_streams,
            commit_empty=commit_empty,
            raise_if_empty=raise_if_empty,
            lock=lock,
        )

    def new_func(*args, **kwargs):
        cmd = Command().command(method)

        if not ignore_std_streams:
            cmd = cmd.track_std_streams()

        # Handle --isolation option:
        if get_git_isolation():
            cmd = cmd.with_git_isolation()

        if clean:
            cmd = cmd.require_clean()

        if requires_migration:
            cmd = cmd.require_migration()

        if commit:
            cmd = cmd.with_commit(commit_only=commit_only, commit_if_empty=commit_empty, raise_if_empty=raise_if_empty)

        if lock or (lock is None and commit):
            cmd = cmd.lock_project()

        cmd.build()

        result = cmd.execute(*args, **kwargs)

        if result.status == CommandResult.FAILURE:
            raise result.error

        return result.output

    return functools.update_wrapper(new_func, method)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test command builder."""

from renku.core.incubation.command import Command, CommandResult


def test_dataset_add_command(project, tmp_path):
    """test dataset add with builder."""
    from renku.core.commands.dataset import DATASET_METADATA_PATHS, _add_to_dataset, create_dataset, list_files

    add_to_dataset = (
        Command()
        .require_clean()
        .require_migration()
        .with_commit(raise_if_empty=True, commit_only=DATASET_METADATA_PATHS)
        .lock_project()
        .command(_add_to_dataset)
        .build()
    )

    create_dataset("ds1", title="", description="", creators=[], commit_message="my awesome dataset")
    data_file = tmp_path / "some-file"
    data_file.write_text("1,2,3", encoding="utf-8")

    result = add_to_dataset.execute([str(data_file)], "ds1")
    assert result.status == CommandResult.SUCCESS
    assert not result.error

    files = list_files(datasets=["ds1"])

    assert isinstance(files, list)
    assert "some-file" in [file_.name for file_ in files]



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service template read manifest controller."""
from marshmallow import EXCLUDE

from renku.core.commands.init import read_template_manifest
from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadOperationMixin
from renku.service.controllers.utils.project_clone import user_project_clone
from renku.service.serializers.templates import ManifestTemplatesRequest, ManifestTemplatesResponseRPC
from renku.service.views import result_response


class TemplatesReadManifestCtrl(ServiceCtrl, ReadOperationMixin):
    """Template read manifest controller."""

    REQUEST_SERIALIZER = ManifestTemplatesRequest()
    RESPONSE_SERIALIZER = ManifestTemplatesResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a templates read manifest controller."""
        self.ctx = TemplatesReadManifestCtrl.REQUEST_SERIALIZER.load({**user_data, **request_data}, unknown=EXCLUDE)
        super(TemplatesReadManifestCtrl, self).__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller context."""
        return self.ctx

    def template_manifest(self):
        """Reads template manifest."""
        project = user_project_clone(self.user_data, self.ctx)
        return read_template_manifest(project.abs_path)

    def renku_op(self):
        """Renku operation for the controller."""
        # NOTE: We leave it empty since it does not execute renku operation.
        pass

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(TemplatesReadManifestCtrl.RESPONSE_SERIALIZER, {"templates": self.template_manifest()})



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Install and uninstall Git hooks.

Prevent modifications of output files
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The commit hooks are enabled by default to prevent situation when
some output file is manually modified.

.. code-block:: console

    $ renku init
    $ renku run echo hello > greeting.txt
    $ edit greeting.txt
    $ git commit greeting.txt
    You are trying to update some output files.

    Modified outputs:
      greeting.txt

    If you are sure, use "git commit --no-verify".

"""

import click

from renku.core.commands.githooks import install_githooks, uninstall_githooks


@click.group()
def githooks():
    """Manage Git hooks for Renku repository."""


@githooks.command()
@click.option("--force", is_flag=True, help="Override existing hooks.")
def install(force):
    """Install Git hooks."""
    install_githooks(force)
    click.secho("OK", fg="green")


@githooks.command()
def uninstall():
    """Uninstall Git hooks."""
    uninstall_githooks()
    click.secho("OK", fg="green")



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service version view tests."""
from renku.core.management.migrate import SUPPORTED_PROJECT_VERSION


def test_version(svc_client):
    """Test expected response from version endpoint."""
    from renku import __version__

    response = svc_client.get("/version")
    assert "result" in response.json
    data = response.json["result"]

    assert {"latest_version", "supported_project_version"} == set(data.keys())
    assert __version__ == data["latest_version"]
    assert SUPPORTED_PROJECT_VERSION == data["supported_project_version"]



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Dataset.

Dataset class allows listing datasets and files inside a Renku project and
accessing their metadata.

To get a list of available datasets in a Renku project use ``list`` method:

.. code-block:: python

    from renku.api import Dataset

    datasets = Dataset.list()

You can then access metadata of a dataset like ``name``, ``title``,
``keywords``, etc. To get the list of files inside a dataset use ``files``
property:

.. code-block:: python

    for dataset_file in dataset.files:
        print(dataset_file.path)

"""

from renku.api.models.project import ensure_project_context
from renku.core.models.datasets import Dataset as CoreDataset


class Dataset:
    """API Dataset model."""

    _ATTRIBUTES = [
        "creators",
        "date_created",
        "date_published",
        "description",
        "keywords",
        "license",
        "name",
        "title",
        "url",
        "version",
    ]

    def __init__(self):
        self._dataset = None

        for name in self._ATTRIBUTES:
            setattr(self, name, None)
        self._files = []

    @classmethod
    def __from_yaml(cls, path, client):
        """Create an instance from Dataset metadata."""
        self = cls()
        self._dataset = CoreDataset.from_yaml(path=path, client=client)
        self._files = [DatasetFile(f) for f in self._dataset.files]

        return self

    @staticmethod
    @ensure_project_context
    def list(project):
        """List all datasets in a project."""
        client = project.client
        return [Dataset.__from_yaml(p, client) for p in client.get_datasets_metadata_files()] if client else []

    def __getattribute__(self, name):
        dataset = object.__getattribute__(self, "_dataset")
        if dataset is not None and name in Dataset._ATTRIBUTES:
            return getattr(dataset, name)

        return object.__getattribute__(self, name)

    @property
    def files(self):
        """Return a list of dataset files."""
        return self._files


class DatasetFile:
    """API DatasetFile model."""

    _ATTRIBUTES = ["added", "full_path", "name", "path"]

    def __init__(self, dataset_file):
        self._dataset_file = dataset_file

        for name in self._ATTRIBUTES:
            setattr(self, name, None)

    def __getattribute__(self, name):
        dataset_file = object.__getattribute__(self, "_dataset_file")
        if dataset_file is not None and name in DatasetFile._ATTRIBUTES:
            return getattr(dataset_file, name)

        return object.__getattribute__(self, name)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Context for jobs."""
import contextlib
import time

from redis import BusyLoadingError

from renku.service.jobs.queues import WorkerQueues


@contextlib.contextmanager
def enqueue_retry(queue, retry=3):
    """Ensure job gets queued."""
    count = 0
    while count < retry:
        try:
            yield WorkerQueues.get(queue)
        except (OSError, IOError, BusyLoadingError):
            time.sleep(2 ** count)
            count += 1
        break



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project clone tests."""
import os
import time
import uuid

import pytest
from flaky import flaky
from marshmallow import EXCLUDE

from renku.service.controllers.utils.project_clone import user_project_clone
from renku.service.serializers.templates import ManifestTemplatesRequest


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
def test_service_user_project_clone(svc_client_cache):
    """Test service user project clone."""
    client, _, cache = svc_client_cache

    user_data = {
        "user_id": uuid.uuid4().hex,
        "email": "contact@justsam.io",
        "fullname": "renku the frog",
        "token": "None",
    }
    project_data = {
        "name": "renku-project-template",
        "depth": 1,
        "url": "https://github.com/SwissDataScienceCenter/renku-project-template",
        "owner": "SwissDataScienceCenter",
    }

    project_data = ManifestTemplatesRequest().load({**user_data, **project_data}, unknown=EXCLUDE)
    project_one = user_project_clone(user_data, project_data)
    time.sleep(1)

    assert project_one.age == 1
    assert not project_one.ttl_expired()
    assert project_one.exists()
    old_path = project_one.abs_path

    os.environ["RENKU_SVC_CLEANUP_TTL_PROJECTS"] = "1"
    time.sleep(1)
    assert project_one.age == 2
    assert project_one.ttl_expired()

    os.environ["RENKU_SVC_CLEANUP_TTL_PROJECTS"] = "3600"
    project_two = user_project_clone(user_data, project_data)
    time.sleep(1)

    assert project_two.age == 1
    assert not project_two.ttl_expired()
    assert project_two.exists()

    new_path = project_two.abs_path
    assert old_path == new_path
    user = cache.get_user(user_data["user_id"])

    projects = [project.project_id for project in cache.get_projects(user)]
    assert project_one.project_id not in projects
    assert project_two.project_id in projects



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test service cache."""
import datetime
import os
import time
import uuid

import pytest


def test_service_cache_ensure_user(svc_client_cache):
    """Test service cache user creation."""
    client, _, cache = svc_client_cache
    expected_id = uuid.uuid4().hex
    user = cache.ensure_user({"user_id": expected_id})

    assert user
    assert expected_id == user.user_id


def test_service_cache_get_users(svc_client_cache):
    """Test getting multiple users."""
    client, _, cache = svc_client_cache

    expected_users = set([cache.ensure_user({"user_id": uuid.uuid4().hex}).user_id for _ in range(10)])

    received_users = set([user.user_id for user in cache.get_users()])

    assert expected_users.issubset(received_users)


def test_service_cache_get_user(svc_client_cache):
    """Test getting a single user."""
    client, _, cache = svc_client_cache
    expected_id = uuid.uuid4().hex
    expected_user = cache.ensure_user({"user_id": expected_id})
    assert expected_user

    received_user = cache.get_user(expected_id)
    assert received_user

    assert expected_user.user_id == received_user.user_id
    assert expected_user.fullname == received_user.fullname
    assert expected_user.email == received_user.email
    assert expected_user.token == received_user.token


def test_service_cache_make_job(svc_client_cache):
    """Test service cache jobs."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": "testuser"})
    job_data = {
        "job_id": uuid.uuid4().hex,
    }

    job = cache.make_job(user, job_data=job_data)

    assert job

    assert job_data["job_id"] == job.job_id
    assert job_data["user_id"] == job.user_id

    assert isinstance(job.created_at, datetime.datetime)
    assert isinstance(job.updated_at, datetime.datetime)

    assert "ENQUEUED" == job.state
    assert job.extras is None


def test_service_cache_get_job(svc_client_cache):
    """Test service get user job."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": "testuser"})
    job_data = {
        "job_id": uuid.uuid4().hex,
    }
    job = cache.make_job(user, job_data=job_data)
    assert job

    retrieved_job = cache.get_job(user, job.job_id)
    assert retrieved_job

    assert job.created_at == retrieved_job.created_at
    assert job.updated_at == retrieved_job.updated_at
    assert job.job_id == retrieved_job.job_id
    assert job.user_id == retrieved_job.user_id
    assert job.state == retrieved_job.state
    assert job.extras == retrieved_job.extras


def test_service_cache_get_jobs(svc_client_cache):
    """Test service get user jobs."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": "testuser1"})
    user2 = cache.ensure_user({"user_id": "testuser2"})

    for _ in range(10):
        job = cache.make_job(user, {"job_id": uuid.uuid4().hex,})
        assert job

    job2 = cache.make_job(user2, {"job_id": uuid.uuid4().hex,})
    assert job2

    retrieved_jobs = cache.get_jobs(user)
    assert retrieved_jobs
    assert 10 == len([job for job in retrieved_jobs])
    assert 1 == len([job for job in cache.get_jobs(user2)])


def test_service_cache_get_job_none(svc_client_cache):
    """Test service get user jobs."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": "testuser"})
    job_data = {
        "job_id": uuid.uuid4().hex,
    }
    job = cache.make_job(user, job_data)

    assert job
    assert cache.get_job(user, None) is None

    with pytest.raises(AttributeError):
        cache.get_job(None, job_data["job_id"])


def test_service_cache_set_file(svc_client_cache):
    """Test service set user file."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": uuid.uuid4().hex})

    file = cache.set_file(
        user, {"file_name": uuid.uuid4().hex, "file_size": 0, "relative_path": "/tmp/renku-core", "is_dir": False}
    )

    assert user.user_id == file.user_id


def test_service_cache_set_files(svc_client_cache):
    """Test service set user files."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": uuid.uuid4().hex})

    files = [
        {"file_name": uuid.uuid4().hex, "file_size": 0, "relative_path": "/tmp/renku-core", "is_dir": False}
        for _ in range(10)
    ]

    expected = set([file_["file_name"] for file_ in files])

    received_files = cache.set_files(user, files)
    received_names = set([file_.file_name for file_ in received_files])

    files_age = set([file_.age for file_ in received_files])
    assert {0} == files_age

    files_ttl_exp = set([file_.ttl_expired() for file_ in received_files])
    assert {False} == files_ttl_exp

    time.sleep(2)
    os.environ["RENKU_SVC_CLEANUP_TTL_FILES"] = "1"
    files_age = set([file_.age for file_ in received_files])
    assert {2} == files_age

    files_ttl_exp = set([file_.ttl_expired() for file_ in received_files])
    assert {True} == files_ttl_exp

    received_users = set([file_.user_id for file_ in received_files])

    assert expected.issubset(received_names)
    assert user.user_id in received_users


def test_service_cache_get_file(svc_client_cache):
    """Test service get user file."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": uuid.uuid4().hex})
    file_obj = cache.set_file(
        user, {"file_name": uuid.uuid4().hex, "file_size": 0, "relative_path": "/tmp/renku-core", "is_dir": False}
    )

    file = cache.get_file(user, file_obj.file_id)

    assert user.user_id == file.user_id


def test_service_cache_get_files(svc_client_cache):
    """Test service get user files."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": uuid.uuid4().hex})

    files_data = [
        {"file_name": uuid.uuid4().hex, "file_size": 0, "relative_path": "/tmp/renku-core", "is_dir": False}
        for _ in range(10)
    ]

    expected_files = {f.file_id for f in cache.set_files(user, files_data)}

    received = [f for f in cache.get_files(user)]
    received_files = {f.file_id for f in received}
    received_user = {f.user_id for f in received}

    assert 10 == len(expected_files)
    assert 10 <= len(received_files)

    assert received_files.issubset(expected_files)
    assert user.user_id in received_user


def test_service_cache_invalidate_file(svc_client_cache):
    """Test service invalidate user file."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": uuid.uuid4().hex})
    file_obj = cache.set_file(
        user, {"file_name": uuid.uuid4().hex, "file_size": 0, "relative_path": "/tmp/renku-core", "is_dir": False}
    )
    cache.invalidate_file(user, file_obj.file_id)

    file = cache.get_file(user, file_obj.file_id)
    assert file is None


def test_service_cache_make_project(svc_client_cache):
    """Test service create project."""
    client, _, cache = svc_client_cache

    user = cache.ensure_user({"user_id": uuid.uuid4().hex})
    project_data = {
        "name": "renku-project-template",
        "depth": 1,
        "git_url": "https://github.com/SwissDataScienceCenter/renku-project-template",
        "email": "contact@justsam.io",
        "fullname": "renku the frog",
        "token": "None",
        "owner": "SwissDataScienceCenter",
    }
    project = cache.make_project(user, project_data)
    time.sleep(1)
    assert project.age == 1
    assert not project.ttl_expired()

    os.environ["RENKU_SVC_CLEANUP_TTL_PROJECTS"] = "1"
    time.sleep(1)
    assert project.age == 2
    assert project.ttl_expired()



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test Python SDK client."""

import inspect

import pytest


def test_local_client(tmpdir):
    """Test a local client."""
    from renku.core.management.client import LocalClient

    client = LocalClient(str(tmpdir.mkdir("project")))

    assert client.path
    assert client.repo is None


@pytest.mark.parametrize(
    "paths, ignored",
    (
        ([".renku.lock"], [".renku.lock"]),
        (["not ignored", "lib/foo", "build/html"], ["lib/foo", "build/html"]),
        (["not ignored"], None),
    ),
)
def test_ignored_paths(paths, ignored, client):
    """Test resolution of ignored paths."""
    assert client.find_ignored_paths(*paths) == ignored


def test_safe_class_attributes(tmpdir):
    """Test that there are no unsafe class attributes on the client.

    This prevents us from adding class attributes that might leak in a threaded environment.
    If you do add a class attribute and want to add it to the list of safe_attributes,
    make sure that it's not something that can leak between calls, e.g. in the service.
    """
    from renku.core.management.client import LocalClient

    # NOTE: attributes that are allowed on LocalClient
    safe_attributes = [
        "ACTIVITY_INDEX",
        "CACHE",
        "CONFIG_NAME",
        "DATASETS",
        "DATA_DIR_CONFIG_KEY",
        "DOCKERFILE",
        "LOCK_SUFFIX",
        "METADATA",
        "POINTERS",
        "RENKU_LFS_IGNORE_PATH",
        "RENKU_PROTECTED_PATHS",
        "TEMPLATE_CHECKSUMS",
        "WORKFLOW",
        "_CMD_STORAGE_CHECKOUT",
        "_CMD_STORAGE_CLEAN",
        "_CMD_STORAGE_INSTALL",
        "_CMD_STORAGE_LIST",
        "_CMD_STORAGE_MIGRATE_INFO",
        "_CMD_STORAGE_PULL",
        "_CMD_STORAGE_STATUS",
        "_CMD_STORAGE_TRACK",
        "_CMD_STORAGE_UNTRACK",
        "_LFS_HEADER",
        "_global_config_dir",
    ]

    client1 = LocalClient(str(tmpdir.mkdir("project1")))

    client2 = LocalClient(str(tmpdir.mkdir("project2")))
    class_attributes = inspect.getmembers(LocalClient, lambda a: not (inspect.isroutine(a)))
    class_attributes = [a for a in class_attributes if not a[0].startswith("__") and not a[0].endswith("__")]
    identical_attributes = []
    for a, v in class_attributes:
        if a in safe_attributes or isinstance(v, property):
            continue
        if id(getattr(client1, a)) == id(getattr(client2, a)):
            identical_attributes.append(a)

    assert not identical_attributes



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command builder."""

import contextlib
import functools
from collections import defaultdict

import click

from renku.core import errors
from renku.core.management import LocalClient
from renku.core.management.config import RENKU_HOME
from renku.core.management.migrate import check_for_migration
from renku.core.management.repository import default_path
from renku.core.utils import communication


def check_finalized(f):
    """Decorator to prevent modification of finalized builders."""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        """Decorator to prevent modification of finalized builders."""
        if not args or not isinstance(args[0], Command):
            raise errors.ParameterError("Command hooks need to be `Command` object methods.")

        if args[0].finalized:
            raise errors.CommandFinalizedError("Cannot modify a finalized `Command`.")

        return f(*args, **kwargs)

    return wrapper


class Command:
    """Base renku command builder."""

    CLIENT_HOOK_ORDER = 1

    def __init__(self):
        """__init__ of Command."""
        self.pre_hooks = defaultdict(list)
        self.post_hooks = defaultdict(list)
        self._operation = None
        self._finalized = False
        self._track_std_streams = False
        self._git_isolation = False

    def __getattr__(self, name):
        """Bubble up attributes of wrapped builders."""
        if "_builder" in self.__dict__:
            return getattr(self._builder, name)

        raise AttributeError(f"{self.__class__.__name__} object has no attribute {name}")

    def __setattr__(self, name, value):
        """Set attributes of wrapped builders."""
        if hasattr(self, "_builder") and self.__class__ is not self._builder.__class__:
            self._builder.__setattr__(name, value)

        object.__setattr__(self, name, value)

    def _pre_hook(self, builder, context, *args, **kwargs):
        """Setup local client."""
        ctx = click.get_current_context(silent=True)
        if ctx is None:
            client = LocalClient(path=default_path(), renku_home=RENKU_HOME, external_storage_requested=True,)
            ctx = click.Context(click.Command(builder._operation))
        else:
            client = ctx.ensure_object(LocalClient)

        stack = contextlib.ExitStack()

        # Handle --isolation option:
        if self._git_isolation:
            client = stack.enter_context(client.worktree())

        context["client"] = client
        context["stack"] = stack
        context["click_context"] = ctx

    def _post_hook(self, builder, context, result, *args, **kwargs):
        """Post-hook method."""

    def execute(self, *args, **kwargs):
        """Execute the wrapped operation.

        First executes `pre_hooks` in ascending `order`, passing a read/write context between them.
        It then calls the wrapped `operation`. The result of the operation then gets pass to all the `post_hooks`,
        also in ascending `order`. It then returns the result or error if there was one.
        """
        if not self.finalized:
            raise errors.CommandNotFinalizedError("Call `build()` before executing a command")

        context = {}
        if any(self.pre_hooks):
            order = sorted(self.pre_hooks.keys())

            for o in order:
                for hook in self.pre_hooks[o]:
                    hook(self, context, *args, **kwargs)

        output = None
        error = None
        try:
            with context["stack"]:
                output = context["click_context"].invoke(self._operation, context["client"], *args, **kwargs)
        except errors.RenkuException as e:
            error = e

        result = CommandResult(output, error, CommandResult.FAILURE if error else CommandResult.SUCCESS)

        if any(self.post_hooks):
            order = sorted(self.post_hooks.keys())

            for o in order:
                for hook in self.post_hooks[o]:
                    hook(self, context, result, *args, **kwargs)

        return result

    @property
    def finalized(self):
        """Whether this builder is still being constructed or has been finalized."""
        if hasattr(self, "_builder"):
            return self._builder.finalized
        return self._finalized

    @check_finalized
    def add_pre_hook(self, order, hook):
        """Add a pre-execution hook.

        :param order: Determines the order of executed hooks, lower numbers get executed first.
        :param hook: The hook to add
        """
        if hasattr(self, "_builder"):
            self._builder.add_pre_hook(order, hook)
        else:
            self.pre_hooks[order].append(hook)

    @check_finalized
    def add_post_hook(self, order, hook):
        """Add a post-execution hook.

        :param order: Determines the order of executed hooks, lower numbers get executed first.
        :param hook: The hook to add
        """
        if hasattr(self, "_builder"):
            self._builder.add_post_hook(order, hook)
        else:
            self.post_hooks[order].append(hook)

    @check_finalized
    def build(self):
        """Build (finalize) the command."""
        if not self._operation:
            raise errors.ConfigurationError("`Command` needs to have a wrapped `command` set")
        self.add_pre_hook(self.CLIENT_HOOK_ORDER, self._pre_hook)

        self._finalized = True

        return self

    @check_finalized
    def command(self, operation):
        """Set the wrapped command.

        :param operation: The function to wrap in the command builder.
        """
        self._operation = operation

        return self

    @check_finalized
    def track_std_streams(self):
        """Whether to track STD streams or not."""
        self._track_std_streams = True

        return self

    @check_finalized
    def with_git_isolation(self):
        """Whether to run in git isolation or not."""
        self._git_isolation = True

        return self

    @check_finalized
    def with_commit(self, message=None, commit_if_empty=False, raise_if_empty=False, commit_only=None):
        """Create a commit.

        :param message: The commit message. Autogenerated if left empty.
        :param commit_if_empty: Whether to commit if there are no modified files .
        :param raise_if_empty: Whether to raise an exception if there are no modified files.
        :param commit_only: Only commit the supplied paths.
        """
        return Commit(self, message, commit_if_empty, raise_if_empty, commit_only)

    @check_finalized
    def lock_project(self):
        """Acquire a lock for the whole project."""
        return ProjectLock(self)

    @check_finalized
    def lock_dataset(self):
        """Acquire a lock for a dataset."""
        return DatasetLock(self)

    @check_finalized
    def require_migration(self):
        """Check if a migration is needed."""
        return RequireMigration(self)

    @check_finalized
    def require_clean(self):
        """Check that the repository is clean."""
        return RequireClean(self)

    @check_finalized
    def with_communicator(self, communicator):
        """Create a communicator."""
        return Communicator(self, communicator)


class Commit(Command):
    """Builder for commands that create a commit."""

    DEFAULT_ORDER = 3

    def __init__(self, builder, message=None, commit_if_empty=False, raise_if_empty=False, commit_only=None):
        """__init__ of Commit.

        :param message: The commit message. Autogenerated if left empty.
        :param commit_if_empty: Whether to commit if there are no modified files .
        :param raise_if_empty: Whether to raise an exception if there are no modified files.
        :param commit_only: Only commit the supplied paths.
        """
        self._builder = builder
        self._message = message
        self._commit_if_empty = commit_if_empty
        self._raise_if_empty = raise_if_empty
        self._commit_filter_paths = commit_only

    def _pre_hook(self, builder, context, *args, **kwargs):
        """Hook to create a commit transaction."""
        if "client" not in context:
            raise ValueError("Commit builder needs a LocalClient to be set.")
        if "stack" not in context:
            raise ValueError("Commit builder needs a stack to be set.")

        transaction = context["client"].transaction(
            clean=False,
            commit=True,
            commit_empty=self._commit_if_empty,
            commit_message=self._message or kwargs.get("commit_message"),
            commit_only=self._commit_filter_paths,
            ignore_std_streams=not builder._track_std_streams,
            raise_if_empty=self._raise_if_empty,
        )
        context["stack"].enter_context(transaction)

    @check_finalized
    def build(self):
        """Build the command."""
        self._builder.add_pre_hook(self.DEFAULT_ORDER, self._pre_hook)

        return self._builder.build()


class RequireMigration(Command):
    """Builder to check for migrations."""

    DEFAULT_ORDER = 2

    def __init__(self, builder):
        """__init__ of RequireMigration."""
        self._builder = builder

    def _pre_hook(self, builder, context, *args, **kwargs):
        """Check if migration is necessary."""
        if "client" not in context:
            raise ValueError("Commit builder needs a LocalClient to be set.")

        check_for_migration(context["client"])

    @check_finalized
    def build(self):
        """Build the command."""
        self._builder.add_pre_hook(self.DEFAULT_ORDER, self._pre_hook)

        return self._builder.build()


class RequireClean(Command):
    """Builder to check if repo is clean."""

    DEFAULT_ORDER = 3

    def __init__(self, builder):
        """__init__ of RequireClean."""
        self._builder = builder

    def _pre_hook(self, builder, context, *args, **kwargs):
        """Check if repo is clean."""
        if "client" not in context:
            raise ValueError("Commit builder needs a LocalClient to be set.")
        context["client"].ensure_clean(ignore_std_streams=not builder._track_std_streams)

    @check_finalized
    def build(self):
        """Build the command."""
        self._builder.add_pre_hook(self.DEFAULT_ORDER, self._pre_hook)

        return self._builder.build()


class ProjectLock(Command):
    """Builder to get a project wide lock."""

    DEFAULT_ORDER = 4

    def __init__(self, builder):
        """__init__ of ProjectLock."""
        self._builder = builder

    def _pre_hook(self, builder, context, *args, **kwargs):
        """Lock the project."""
        if "client" not in context:
            raise ValueError("Commit builder needs a LocalClient to be set.")
        if "stack" not in context:
            raise ValueError("Commit builder needs a stack to be set.")

        context["stack"].enter_context(context["client"].lock)

    @check_finalized
    def build(self):
        """Build the command."""
        self._builder.add_pre_hook(self.DEFAULT_ORDER, self._pre_hook)

        return self._builder.build()


class DatasetLock(Command):
    """Builder to lock on a dataset."""

    DEFAULT_ORDER = 5

    def __init__(self, builder):
        """__init__ of DatasetLock."""
        self._builder = builder

    def _pre_hook(self, builder, context, *args, **kwargs):
        raise NotImplementedError()

    @check_finalized
    def build(self):
        """Build the command."""
        self._builder.add_pre_hook(self.DEFAULT_ORDER, self._pre_hook)

        return self._builder.build()


class Communicator(Command):
    """Hook for logging and interaction with user."""

    DEFAULT_ORDER = 6

    def __init__(self, builder, communicator):
        """__init__ of Communicator.

        :param communicator: Instance of CommunicationCallback.
        """
        self._builder = builder
        self._communicator = communicator

    def _pre_hook(self, builder, context, *args, **kwargs):
        communication.subscribe(self._communicator)

    def _post_hook(self, builder, context, result, *args, **kwargs):
        communication.unsubscribe(self._communicator)

    @check_finalized
    def build(self):
        """Build the command."""
        self._builder.add_pre_hook(self.DEFAULT_ORDER, self._pre_hook)
        self._builder.add_post_hook(self.DEFAULT_ORDER, self._post_hook)

        return self._builder.build()


class CommandResult:
    """The result of a command.

    The return value of the command is set as `.output`, if there was an error, it is set as `.error`, and
    the status of the command is set to either `CommandResult.SUCCESS` or CommandResult.FAILURE`.
    """

    SUCCESS = 0

    FAILURE = 1

    def __init__(self, output, error, status):
        """__init__ of CommandResult."""
        self.output = output
        self.error = error
        self.status = status



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku migrations management.

Migrations files are put in renku/core/management/migrations directory. Name
of these files has m_1234__name.py format where 1234 is the migration version
and name can be any alphanumeric and underscore combination. Migration files
are sorted based on their lowercase name. Each migration file must define a
public "migrate" function that accepts a client as its argument.

When executing a migration, the migration file is imported as a module and the
"migrate" function is executed. Migration version is checked against the Renku
project version (in .renku/metadata.yml) and any migration which has a higher
version is applied to the project.
"""
import hashlib
import importlib
import json
import os
import re
import shutil
from pathlib import Path

import pkg_resources
from jinja2 import Template

from renku.core.errors import (
    DockerfileUpdateError,
    MigrationError,
    MigrationRequired,
    ProjectNotSupported,
    TemplateUpdateError,
)
from renku.core.utils.migrate import read_project_version

SUPPORTED_PROJECT_VERSION = 8


def check_for_migration(client):
    """Checks if migration is required."""
    if is_migration_required(client):
        raise MigrationRequired
    elif is_project_unsupported(client):
        raise ProjectNotSupported


def is_migration_required(client):
    """Check if project requires migration."""
    return is_renku_project(client) and _get_project_version(client) < SUPPORTED_PROJECT_VERSION


def is_project_unsupported(client):
    """Check if this version of Renku cannot work with the project."""
    return is_renku_project(client) and _get_project_version(client) > SUPPORTED_PROJECT_VERSION


def is_template_update_possible(client):
    """Check if the project can be updated to a newer version of the project template."""
    return _update_template(client, check_only=True)


def is_docker_update_possible(client):
    """Check if the Dockerfile can be updated to a new version of renku-python."""
    return _update_dockerfile(client, check_only=True)


def migrate(
    client,
    force_template_update=False,
    skip_template_update=False,
    skip_docker_update=False,
    skip_migrations=False,
    progress_callback=None,
):
    """Apply all migration files to the project."""
    template_updated = docker_updated = False
    if not is_renku_project(client):
        return False, template_updated, docker_updated

    if (
        not skip_template_update
        and client.project.template_source
        and (force_template_update or client.project.automated_update)
    ):
        try:
            template_updated, _, _ = _update_template(client, progress_callback=progress_callback)
        except TemplateUpdateError:
            raise
        except (Exception, BaseException) as e:
            raise TemplateUpdateError("Couldn't update from template.") from e

    if not skip_docker_update:
        try:
            docker_updated = _update_dockerfile(client, progress_callback=progress_callback)
        except DockerfileUpdateError:
            raise
        except (Exception, BaseException) as e:
            raise DockerfileUpdateError("Couldn't update renku version in Dockerfile.") from e

    if skip_migrations:
        return False, template_updated, docker_updated

    project_version = _get_project_version(client)
    n_migrations_executed = 0

    for version, path in get_migrations():
        if version > project_version:
            module = importlib.import_module(path)
            if progress_callback:
                module_name = module.__name__.split(".")[-1]
                progress_callback(f"Applying migration {module_name}...")
            try:
                module.migrate(client)
            except (Exception, BaseException) as e:
                raise MigrationError("Couldn't execute migration") from e
            n_migrations_executed += 1
    if n_migrations_executed > 0:
        client._project = None  # NOTE: force reloading of project metadata
        client.project.version = str(version)
        client.project.to_yaml()

        if progress_callback:
            progress_callback(f"Successfully applied {n_migrations_executed} migrations.")

    return n_migrations_executed != 0, template_updated, docker_updated


def _update_template(client, check_only=False, progress_callback=None):
    """Update local files from the remote template."""
    from renku.core.commands.init import fetch_template

    project = client.project

    if not project.template_version:
        return False, None, None

    template_manifest, template_folder, template_source, template_version = fetch_template(
        project.template_source, project.template_ref, progress_callback
    )

    if template_source == "renku":
        template_version = pkg_resources.parse_version(template_version)
        current_version = pkg_resources.parse_version(project.template_version)
        if template_version <= current_version:
            return False, project.template_version, current_version
    else:
        if template_version == project.template_version:
            return False, project.template_version, template_version

    if check_only:
        return True, project.template_version, template_version

    if progress_callback:
        progress_callback("Updating project from template...")

    template_filtered = [
        template_elem for template_elem in template_manifest if template_elem["folder"] == project.template_id
    ]
    if len(template_filtered) == 1:
        template_data = template_filtered[0]
    else:
        raise TemplateUpdateError(f'The template with id "{project.template_id}" is not available.')

    template_path = template_folder / template_data["folder"]

    metadata = json.loads(project.template_metadata)

    template_variables = set(template_data.get("variables", {}).keys())
    metadata_keys = set(metadata.keys())
    missing_keys = ", ".join(template_variables - metadata_keys)
    if missing_keys:
        raise TemplateUpdateError(
            f"Can't update template, it now requires variable(s) {missing_keys} which were not present on project "
            "creation."
        )

    if not os.path.exists(client.template_checksums):
        raise TemplateUpdateError("Can't update template as there are no template checksums set on the project.")

    with open(client.template_checksums, "r") as checksum_file:
        checksums = json.load(checksum_file)

    updated_files = []

    for file in template_path.glob("**/*"):
        rel_path = file.relative_to(template_path)
        destination = client.path / rel_path

        # NOTE: the path could contain template variables, we need to template it
        destination = Path(Template(str(destination)).render(metadata))

        try:
            # parse file and process it
            template = Template(file.read_text())
            rendered_content = template.render(metadata)
            sha256_hash = hashlib.sha256()
            content_bytes = rendered_content.encode("utf-8")
            blocksize = 4096
            blocks = (len(content_bytes) - 1) // blocksize + 1
            for i in range(blocks):
                byte_block = content_bytes[i * blocksize : (i + 1) * blocksize]
                sha256_hash.update(byte_block)
            new_template_hash = sha256_hash.hexdigest()

            if not destination.exists() and str(rel_path) not in checksums:
                # NOTE: new file in template
                local_changes = False
                remote_changes = True
            else:
                current_hash = None  # NOTE: None if user deleted file locally

                if destination.exists():
                    current_hash = client._content_hash(destination)

                local_changes = current_hash != checksums[str(rel_path)]
                remote_changes = new_template_hash != checksums[str(rel_path)]

            if local_changes:
                if remote_changes and str(rel_path) in project.immutable_template_files:
                    # NOTE: There are local changes in a file that should not be changed by users,
                    # and the file was updated in the template as well. So the template can't be updated.
                    raise TemplateUpdateError(
                        f"Can't update template as immutable template file {rel_path} has local changes."
                    )
                continue
            elif not remote_changes:
                continue

            destination.write_text(rendered_content)
        except IsADirectoryError:
            destination.mkdir(parents=True, exist_ok=True)
        except TypeError:
            shutil.copy(file, destination)

    if progress_callback:
        updated = "\n".join(updated_files)
        progress_callback(f"Updated project from template, updated files:\n{updated}")

    project.template_version = template_version
    project.to_yaml()

    return True, project.template_version, template_version


def _update_dockerfile(client, check_only=False, progress_callback=None):
    """Update the dockerfile to the newest version of renku."""
    from renku import __version__

    if not client.docker_path.exists():
        return False

    if progress_callback:
        progress_callback("Updating dockerfile...")

    with open(client.docker_path, "r") as f:
        dockercontent = f.read()

    current_version = pkg_resources.parse_version(__version__)
    m = re.search(r"^ARG RENKU_VERSION=(\d+\.\d+\.\d+)$", dockercontent, flags=re.MULTILINE)
    if not m:
        if check_only:
            return False
        raise DockerfileUpdateError(
            "Couldn't update renku-python version in Dockerfile, as it doesn't contain an 'ARG RENKU_VERSION=...' line."
        )

    docker_version = pkg_resources.parse_version(m.group(1))

    if docker_version >= current_version:
        return False

    if check_only:
        return True

    dockercontent = re.sub(
        r"^ARG RENKU_VERSION=\d+\.\d+\.\d+$", f"ARG RENKU_VERSION={__version__}", dockercontent, flags=re.MULTILINE,
    )

    with open(client.docker_path, "w") as f:
        f.write(dockercontent)

    if progress_callback:
        progress_callback("Updated dockerfile.")

    return True


def _get_project_version(client):
    try:
        return int(read_project_version(client))
    except ValueError:
        return 1


def is_renku_project(client):
    """Check if repository is a renku project."""
    try:
        return client.project is not None
    except ValueError:  # Error in loading due to an older schema
        return client.renku_metadata_path.exists()


def get_migrations():
    """Return a sorted list of versions and migration modules."""
    migrations = []
    for file_ in pkg_resources.resource_listdir("renku.core.management", "migrations"):
        match = re.search(r"m_([0-9]{4})__[a-zA-Z0-9_-]*.py", file_)

        if match is None:  # migration files match m_0000__[name].py format
            continue

        version = int(match.groups()[0])
        path = "renku.core.management.migrations.{}".format(Path(file_).stem)
        migrations.append((version, path))

    migrations = sorted(migrations, key=lambda v: v[1].lower())
    return migrations



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check exceptions raised on views."""
import json
import uuid

import pytest
from flaky import flaky

from renku.service.config import INVALID_HEADERS_ERROR_CODE


@pytest.mark.service
def test_allowed_methods_exc(service_allowed_endpoint):
    """Check allowed methods for every endpoint."""
    methods, request, svc_client = service_allowed_endpoint

    method = request["allowed_method"]
    if method == "GET":  # if GET remove sister method HEAD
        methods.pop(method)
        methods.pop("HEAD")
    else:
        methods.pop(method)

    for method, fn in methods.items():
        response = fn(request["url"])
        assert 405 == response.status_code


@pytest.mark.service
def test_auth_headers_exc(service_allowed_endpoint):
    """Check correct headers for every endpoint."""
    methods, request, svc_client = service_allowed_endpoint

    method = request["allowed_method"]
    if method == "GET":  # if GET remove sister method HEAD
        client_method = methods.pop(method)
        methods.pop("HEAD")
    else:
        client_method = methods.pop(method)

    response = client_method(request["url"], headers=request["headers"],)

    assert 200 == response.status_code
    assert INVALID_HEADERS_ERROR_CODE == response.json["error"]["code"]

    err_message = "user identification is incorrect or missing"
    assert err_message == response.json["error"]["reason"]


@pytest.mark.service
@pytest.mark.integration
@flaky(max_runs=30, min_passes=1)
def test_migration_required_flag(svc_client_setup):
    """Check migration required failure."""
    svc_client, headers, project_id, _ = svc_client_setup

    payload = {
        "project_id": project_id,
        "name": "{0}".format(uuid.uuid4().hex),
    }

    response = svc_client.post("/datasets.create", data=json.dumps(payload), headers=headers,)

    assert response.json["error"]["migration_required"]


@pytest.mark.service
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
def test_project_uninitialized(svc_client, it_remote_non_renku_repo, authentication_headers):
    """Check migration required failure."""
    payload = {"git_url": it_remote_non_renku_repo}

    response = svc_client.post("/cache.project_clone", data=json.dumps(payload), headers=authentication_headers,)

    assert response
    assert "result" in response.json
    assert "error" not in response.json

    project_id = response.json["result"]["project_id"]
    initialized = response.json["result"]["initialized"]

    assert not initialized

    payload = {
        "project_id": project_id,
        "name": "{0}".format(uuid.uuid4().hex),
    }

    response = svc_client.post("/datasets.create", data=json.dumps(payload), headers=authentication_headers,)

    assert response
    assert "error" in response.json
    assert "project_initialization_required" in response.json["error"]
    assert response.json["error"]["project_initialization_required"]


@pytest.mark.service
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize(
    "git_url,expected",
    [
        ("https://github.com", {"error": {"code": -32602, "reason": "Validation error: `schema` - Invalid `git_url`"}}),
        (
            "https://github.com/SwissDataScienceCenter",
            {"error": {"code": -32602, "reason": "Validation error: `schema` - Invalid `git_url`"}},
        ),
        ("https://test.com/test2/test3", {"error": {"code": -32001, "reason": "Repository could not be found"}}),
        ("https://www.test.com/test2/test3", {"error": {"code": -32001, "reason": "Repository could not be found"}}),
    ],
)
def test_invalid_git_remote(git_url, expected, svc_client_with_templates):
    """Check reading manifest template."""
    svc_client, headers, template_params = svc_client_with_templates
    template_params["url"] = git_url
    response = svc_client.get("/templates.read_manifest", query_string=template_params, headers=headers)

    assert response
    assert expected == response.json



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Project.

Project class acts as a context for other Renku entities like Dataset, or
Inputs/Outputs. It provides access to internals of a Renku project for such
entities.

Normally, you do not need to create an instance of Project class directly
unless you want to have access to Project metadata (e.g. path). To separate
parts of your script that uses Renku entities, you can create a Project context
manager and interact with Renku inside it:

.. code-block:: python

    from renku.api import Project, Input

    with Project():
        input_1 = Input("data_1")

"""
from functools import wraps

from git import GitError, Repo
from werkzeug.local import LocalStack


class Project:
    """API Project context class."""

    _project_contexts = LocalStack()

    def __init__(self):
        self._client = _get_local_client()

    def __enter__(self):
        self._project_contexts.push(self)

        return self

    def __exit__(self, type, value, traceback):
        project_context = self._project_contexts.pop()
        if project_context is not self:
            raise RuntimeError("Project context was changed.")

    @property
    def client(self):
        """Return the LocalClient instance."""
        return self._client

    @property
    def path(self):
        """Absolute path to project's root directory."""
        return self._client.path.resolve()


def ensure_project_context(fn):
    """Check existence of a project context."""

    @wraps(fn)
    def wrapper(*args, **kwargs):
        project = _get_current_project() or Project()
        return fn(*args, **kwargs, project=project)

    return wrapper


def _get_current_project():
    """Return current project context if any or a new project object."""
    return Project._project_contexts.top if Project._project_contexts.top else None


def _get_local_client():
    from renku.core.management.client import LocalClient

    try:
        repo = Repo(".", search_parent_directories=True)
    except GitError:
        path = "."
    else:
        path = repo.working_dir

    return LocalClient(path)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates view."""
from flask import Blueprint, request
from flask_apispec import marshal_with, use_kwargs

from renku.service.config import SERVICE_PREFIX
from renku.service.controllers.templates_create_project import TemplatesCreateProjectCtrl
from renku.service.controllers.templates_read_manifest import TemplatesReadManifestCtrl
from renku.service.serializers.templates import (
    ManifestTemplatesRequest,
    ManifestTemplatesResponseRPC,
    ProjectTemplateRequest,
    ProjectTemplateResponseRPC,
)
from renku.service.views.decorators import (
    accepts_json,
    handle_common_except,
    header_doc,
    requires_cache,
    requires_identity,
)

TEMPLATES_BLUEPRINT_TAG = "templates"
templates_blueprint = Blueprint(TEMPLATES_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX)


@use_kwargs(ManifestTemplatesRequest, location="query")
@marshal_with(ManifestTemplatesResponseRPC)
@header_doc("Clone a remote template repository and read the templates.", tags=(TEMPLATES_BLUEPRINT_TAG,))
@templates_blueprint.route(
    "/templates.read_manifest", methods=["GET"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def read_manifest_from_template(user_data, cache):
    """Read templates from the manifest file of a template repository."""
    return TemplatesReadManifestCtrl(cache, user_data, dict(request.args)).to_response()


@use_kwargs(ProjectTemplateRequest)
@marshal_with(ProjectTemplateResponseRPC)
@header_doc(
    "Create a new project starting from a target template available in a " "remote repositpry.",
    tags=(TEMPLATES_BLUEPRINT_TAG,),
)
@templates_blueprint.route(
    "/templates.create_project", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def create_project_from_template(user_data, cache):
    """Create a new project starting form target template."""
    return TemplatesCreateProjectCtrl(cache, user_data, dict(request.json)).to_response()



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``config`` command."""
import pytest

from renku.cli import cli


def test_config_value_locally(client, runner, project, global_config_dir):
    """Check setting/getting from local configuration."""
    result = runner.invoke(cli, ["config", "set", "key", "local-value"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["config", "show", "key"])
    assert 0 == result.exit_code
    assert result.output == "local-value\n"
    # Value set locally is not visible globally
    result = runner.invoke(cli, ["config", "show", "key", "--global"])
    assert 2 == result.exit_code

    # Reading non-existing values is an error
    result = runner.invoke(cli, ["config", "show", "non-existing"])
    assert 2 == result.exit_code


def test_config_value_globally(client, runner, project, global_config_dir):
    """Check setting/getting from global configuration."""
    result = runner.invoke(cli, ["config", "set", "key", "global-value", "--global"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["config", "show", "key"])
    assert 0 == result.exit_code
    assert result.output == "global-value\n"
    result = runner.invoke(cli, ["config", "show", "key", "--global"])
    assert 0 == result.exit_code
    assert result.output == "global-value\n"
    # Value set globally is not visible in local config
    result = runner.invoke(cli, ["config", "show", "key", "--local"])
    assert 2 == result.exit_code


def test_config_get_non_existing_value(client, runner, project, global_config_dir):
    """Check getting non-existing value is an error."""
    result = runner.invoke(cli, ["config", "show", "non-existing"])
    assert 2 == result.exit_code


def test_local_overrides_global_config(client, runner, project, global_config_dir):
    """Test setting config both global and locally."""
    result = runner.invoke(cli, ["config", "set", "key", "global-value", "--global"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["config", "show", "key"])
    assert 0 == result.exit_code
    assert result.output == "global-value\n"

    result = runner.invoke(cli, ["config", "set", "key", "local-value"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["config", "show", "key"])
    assert 0 == result.exit_code
    assert result.output == "local-value\n"


@pytest.mark.parametrize("global_only", (False, True))
def test_config_remove_value_locally(client, runner, project, global_config_dir, global_only):
    """Check removing value from local configuration."""
    param = ["--global"] if global_only else []
    result = runner.invoke(cli, ["config", "set", "key", "some-value"] + param)
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["config", "show", "key"] + param)
    assert "some-value\n" == result.output

    result = runner.invoke(cli, ["config", "remove", "key"] + param)
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["config", "show", "key"] + param)
    assert "some-value" not in result.output


def test_local_config_committed(client, runner, data_repository, global_config_dir):
    """Test local configuration update is committed only when it is changed."""
    commit_sha_before = client.repo.head.object.hexsha

    result = runner.invoke(cli, ["config", "set", "local-key", "value"])
    assert 0 == result.exit_code
    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_after != commit_sha_before

    # Adding the same config should not create a new commit
    commit_sha_before = client.repo.head.object.hexsha

    result = runner.invoke(cli, ["config", "set", "local-key", "value"])
    assert 0 == result.exit_code
    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_after == commit_sha_before

    # Adding a global config should not create a new commit
    result = runner.invoke(cli, ["config", "set", "global-key", "value", "--global"])
    assert 0 == result.exit_code
    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_after == commit_sha_before


@pytest.mark.parametrize(
    "args,message", [(["show", "--local", "--global", "key"], "Cannot use --local and --global together."),],
)
def test_invalid_command_args(client, runner, project, global_config_dir, args, message):
    """Test invalid combination of command-line arguments."""
    result = runner.invoke(cli, ["config"] + args)
    assert 2 == result.exit_code
    assert message in result.output


@pytest.mark.parametrize("config_key", ["data_directory"])
def test_readonly_config(client, runner, project, config_key):
    """Test readonly config can only be set once."""
    result = runner.invoke(cli, ["config", "set", config_key, "value"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["config", "set", config_key, "value"])
    assert 2 == result.exit_code
    assert f"Configuration {config_key} cannot be modified." in result.output

    result = runner.invoke(cli, ["config", "remove", config_key])
    assert 2 == result.exit_code
    assert f"Configuration {config_key} cannot be modified." in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate project to the latest Renku version."""
import os

import click

from renku.core.commands.client import pass_local_client
from renku.core.commands.echo import WARNING
from renku.core.commands.migrate import (
    MIGRATION_REQUIRED,
    NON_RENKU_REPOSITORY,
    UNSUPPORTED_PROJECT,
    check_project,
    migrate_project,
    migrate_project_no_commit,
)
from renku.core.errors import MigrationRequired, ProjectNotSupported


@click.command()
@click.option("-c", "--check", is_flag=True, help="Check if migration is required and quit.")
@click.option("--no-commit", is_flag=True, hidden=True, help="Do not commit changes after the migration.")
def migrate(check, no_commit):
    """Check for migration and migrate to the latest Renku project version."""
    if check:
        status = check_project()
        if status == MIGRATION_REQUIRED:
            raise MigrationRequired
        elif status == UNSUPPORTED_PROJECT:
            raise ProjectNotSupported
        elif status == NON_RENKU_REPOSITORY:
            click.secho(WARNING + "Not a renku project.")

        return

    if no_commit:
        result, _, _ = migrate_project_no_commit(
            skip_template_update=True, skip_docker_update=True, progress_callback=click.secho,
        )
    else:
        result, _, _ = migrate_project(
            skip_template_update=True, skip_docker_update=True, progress_callback=click.secho,
        )

    if result:
        click.secho("OK", fg="green")
    else:
        if check_project() == NON_RENKU_REPOSITORY:
            click.secho(WARNING + "Not a renku project.")
        click.secho("No migrations required.")


@click.command(hidden=True)
@click.argument(
    "paths", type=click.Path(exists=True, dir_okay=True), nargs=-1, required=True,
)
@pass_local_client
def check_immutable_template_files(client, paths):
    """Check specified paths if they are marked immutable in the template."""
    paths = client.check_immutable_template_files(*paths)
    if paths:
        click.echo(os.linesep.join(paths))
        exit(1)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent an annotation for a workflow."""

import attr
from marshmallow import EXCLUDE

from renku.core.models.calamus import JsonLDSchema, dcterms, fields, oa


@attr.s(cmp=False,)
class Annotation:
    """Represents a custom annotation for a research object."""

    _id = attr.ib(kw_only=True)

    body = attr.ib(default=None, kw_only=True)

    source = attr.ib(default=None, kw_only=True)

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return AnnotationSchema().load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return AnnotationSchema().dump(self)


class AnnotationSchema(JsonLDSchema):
    """Annotation schema."""

    class Meta:
        """Meta class."""

        rdf_type = oa.Annotation
        model = Annotation
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    body = fields.Raw(oa.hasBody)
    source = fields.Raw(dcterms.creator)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for dataset tags list."""

from collections import OrderedDict

from renku.core.models.tabulate import tabulate


def tabular(client, tags):
    """Format dataset tags with a tabular output.

    :param client: LocalClient instance.
    :param tags: Dataset tags.
    """
    return tabulate(
        tags,
        headers=OrderedDict(
            (("created", None), ("name", None), ("description", None), ("dataset", None), ("commit", None),)
        ),
        # workaround for tabulate issue 181
        # https://bitbucket.org/astanin/python-tabulate/issues/181/disable_numparse-fails-on-empty-input
        disable_numparse=[1, 2, 4] if len(tags) > 0 else False,
    )


def jsonld(client, tags):
    """Format dataset tags as JSON-LD.

    :param client: LocalClient instance.
    :param tags: Dataset tags.
    """
    from renku.core.models.json import dumps

    data = [tag.as_jsonld() for tag in tags]
    return dumps(data, indent=2)


DATASET_TAGS_FORMATS = {
    "tabular": tabular,
    "json-ld": jsonld,
}
"""Valid formatting options."""



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets view."""
from flask import Blueprint, request
from flask_apispec import marshal_with, use_kwargs

from renku.service.config import SERVICE_PREFIX
from renku.service.controllers.datasets_add_file import DatasetsAddFileCtrl
from renku.service.controllers.datasets_create import DatasetsCreateCtrl
from renku.service.controllers.datasets_edit import DatasetsEditCtrl
from renku.service.controllers.datasets_files_list import DatasetsFilesListCtrl
from renku.service.controllers.datasets_import import DatasetsImportCtrl
from renku.service.controllers.datasets_list import DatasetsListCtrl
from renku.service.controllers.datasets_remove import DatasetsRemoveCtrl
from renku.service.controllers.datasets_unlink import DatasetsUnlinkCtrl
from renku.service.serializers.datasets import (
    DatasetAddRequest,
    DatasetAddResponseRPC,
    DatasetCreateRequest,
    DatasetCreateResponseRPC,
    DatasetEditRequest,
    DatasetEditResponseRPC,
    DatasetFilesListRequest,
    DatasetFilesListResponseRPC,
    DatasetImportRequest,
    DatasetImportResponseRPC,
    DatasetListRequest,
    DatasetListResponseRPC,
    DatasetRemoveRequest,
    DatasetRemoveResponseRPC,
    DatasetUnlinkRequest,
    DatasetUnlinkResponseRPC,
)
from renku.service.views.decorators import (
    accepts_json,
    handle_common_except,
    header_doc,
    requires_cache,
    requires_identity,
)

DATASET_BLUEPRINT_TAG = "datasets"
dataset_blueprint = Blueprint(DATASET_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX)


@use_kwargs(DatasetListRequest, location="query")
@marshal_with(DatasetListResponseRPC)
@header_doc("List all datasets in project.", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.list", methods=["GET"], provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_datasets_view(user_data, cache):
    """List all datasets in project."""
    return DatasetsListCtrl(cache, user_data, dict(request.args)).to_response()


@use_kwargs(DatasetFilesListRequest, location="query")
@marshal_with(DatasetFilesListResponseRPC)
@header_doc("List files in a dataset.", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.files_list", methods=["GET"], provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_dataset_files_view(user_data, cache):
    """List files in a dataset."""
    return DatasetsFilesListCtrl(cache, user_data, dict(request.args)).to_response()


@use_kwargs(DatasetAddRequest)
@marshal_with(DatasetAddResponseRPC)
@header_doc("Add an uploaded file to a cloned repository.", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.add", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def add_file_to_dataset_view(user_data, cache):
    """Add the uploaded file to cloned repository."""
    return DatasetsAddFileCtrl(cache, user_data, dict(request.json)).to_response()


@use_kwargs(DatasetCreateRequest)
@marshal_with(DatasetCreateResponseRPC)
@header_doc("Create a new dataset in a project.", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.create", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def create_dataset_view(user_data, cache):
    """Create a new dataset in a project."""
    return DatasetsCreateCtrl(cache, user_data, dict(request.json)).to_response()


@use_kwargs(DatasetRemoveRequest)
@marshal_with(DatasetRemoveResponseRPC)
@header_doc("Remove a dataset from a project.", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.remove", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def remove_dataset_view(user_data, cache):
    """Remove a dataset from a project."""
    return DatasetsRemoveCtrl(cache, user_data, dict(request.json)).to_response()


@use_kwargs(DatasetImportRequest)
@marshal_with(DatasetImportResponseRPC)
@header_doc("Import a dataset", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.import", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def import_dataset_view(user_data, cache):
    """Import a dataset view."""
    return DatasetsImportCtrl(cache, user_data, dict(request.json)).to_response()


@use_kwargs(DatasetEditRequest)
@marshal_with(DatasetEditResponseRPC)
@header_doc("Edit dataset metadata", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.edit", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def edit_dataset_view(user_data, cache):
    """Edit dataset metadata."""
    return DatasetsEditCtrl(cache, user_data, dict(request.json)).to_response()


@use_kwargs(DatasetUnlinkRequest)
@marshal_with(DatasetUnlinkResponseRPC)
@header_doc("Unlink a file from a dataset", tags=(DATASET_BLUEPRINT_TAG,))
@dataset_blueprint.route(
    "/datasets.unlink", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def unlink_file_view(user_data, cache):
    """Unlink a file from a dataset."""
    return DatasetsUnlinkCtrl(cache, user_data, dict(request.json)).to_response()



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initial migrations."""

import glob
import os
import uuid
from functools import cmp_to_key
from hashlib import sha1
from pathlib import Path

from cwlgen import CommandLineTool, parse_cwl
from cwlgen.requirements import InitialWorkDirRequirement
from git import NULL_TREE, Actor
from werkzeug.utils import secure_filename

from renku.core.management.migrations.models.v3 import Dataset
from renku.core.models.entities import Collection, Entity
from renku.core.models.provenance.activities import ProcessRun, WorkflowRun
from renku.core.models.provenance.agents import Person, SoftwareAgent
from renku.core.models.workflow.parameters import CommandArgument, CommandInput, CommandOutput, MappedIOStream
from renku.core.models.workflow.run import Run
from renku.core.utils.scm import git_unicode_unescape
from renku.version import __version__, version_url

default_missing_software_agent = SoftwareAgent(
    label="renku pre 0.11.0", id="https://github.com/swissdatasciencecenter/renku-python/tree/pre-0.11.0"
)


def migrate(client):
    """Migration function."""
    _migrate_old_workflows(client)


def _migrate_old_workflows(client):
    """Migrates old cwl workflows to new jsonld format."""
    wf_path = "{}/*.cwl".format(client.workflow_path)

    cwl_paths = glob.glob(wf_path)

    cwl_paths = [(p, _find_only_cwl_commit(client, p)) for p in cwl_paths]

    def _sort_cwl_commits(e1, e2):
        """Sorts cwl commits in order of their creation."""
        commit1 = e1[1]
        commit2 = e2[1]

        if client.repo.is_ancestor(commit1, commit2):
            return -1
        if client.repo.is_ancestor(commit2, commit1):
            return 1

        if commit1.committed_date < commit2.committed_date:
            return -1
        if commit1.committed_date > commit2.committed_date:
            return 1

        if commit1.authored_date < commit2.authored_date:
            return -1
        if commit1.authored_date > commit2.authored_date:
            return 1
        raise ValueError(
            f"Cannot order commits {commit1} and {commit2}, there is no "
            "dependency between them and they have identical commit and "
            "author dates"
        )

    cwl_paths = sorted(cwl_paths, key=cmp_to_key(_sort_cwl_commits))

    for cwl_file, commit in cwl_paths:
        path = _migrate_cwl(client, cwl_file, commit)
        os.remove(cwl_file)

        client.repo.git.add(cwl_file, path)

        if client.repo.is_dirty():
            commit_msg = "renku migrate: " "committing migrated workflow"

            committer = Actor("renku {0}".format(__version__), version_url)

            client.repo.index.commit(
                commit_msg, committer=committer, skip_hooks=True,
            )


def _migrate_cwl(client, path, commit):
    """Migrate a cwl file."""
    workflow = parse_cwl_cached(str(path))

    if isinstance(workflow, CommandLineTool):
        _, path = _migrate_single_step(client, workflow, path, commit=commit, persist=True)
    else:
        _, path = _migrate_composite_step(client, workflow, path, commit=commit)

    return path


def _migrate_single_step(client, cmd_line_tool, path, commit=None, parent_commit=None, persist=False):
    """Migrate a single step workflow."""
    if not commit:
        commit = client.find_previous_commit(path, revision=parent_commit if parent_commit else "HEAD")

    run = Run(client=client, path=path, commit=commit)
    run.command = " ".join(cmd_line_tool.baseCommand)
    run.successcodes = cmd_line_tool.successCodes

    inputs = list(cmd_line_tool.inputs)
    outputs = list(cmd_line_tool.outputs)

    # NOTE: Make run ids deterministic to prevent duplication.
    rel_path = Path(path).relative_to(client.path)
    if parent_commit:
        label = f"{rel_path}@{parent_commit.hexsha}"
    else:
        label = f"{rel_path}@{commit.hexsha}"
    identifier = sha1(label.encode("utf-8")).hexdigest()

    base_id = Run.generate_id(client, identifier=identifier)
    run._id = base_id

    if cmd_line_tool.stdin:
        name = cmd_line_tool.stdin.split(".")[1]

        if name.endswith(")"):
            name = name[:-1]

        matched_input = next(i for i in inputs if i.id == name)
        inputs.remove(matched_input)

        path = client.workflow_path / Path(matched_input.default["path"])
        stdin = path.resolve().relative_to(client.path)
        id_ = CommandInput.generate_id(base_id, "stdin")

        run.inputs.append(
            CommandInput(
                id=id_,
                consumes=_entity_from_path(client, stdin, commit),
                mapped_to=MappedIOStream(client=client, stream_type="stdin"),
            )
        )

    if cmd_line_tool.stdout:
        run.outputs.append(
            CommandOutput(
                id=CommandOutput.generate_id(base_id, "stdout"),
                produces=_entity_from_path(client, cmd_line_tool.stdout, commit),
                mapped_to=MappedIOStream(client=client, stream_type="stdout"),
                create_folder=False,
            )
        )

        matched_output = next(o for o in outputs if o.id == "output_stdout")

        if matched_output:
            outputs.remove(matched_output)

    if cmd_line_tool.stderr:
        run.outputs.append(
            CommandOutput(
                id=CommandOutput.generate_id(base_id, "stderr"),
                produces=_entity_from_path(client, cmd_line_tool.stderr, commit),
                mapped_to=MappedIOStream(client=client, stream_type="stderr"),
                create_folder=False,
            )
        )

        matched_output = next(o for o in outputs if o.id == "output_stderr")

        if matched_output:
            outputs.remove(matched_output)

    created_outputs = []
    workdir_requirements = [r for r in cmd_line_tool.requirements if isinstance(r, InitialWorkDirRequirement)]

    for r in workdir_requirements:
        for listing in r.listing:
            if listing.entry == '$({"listing": [], "class": "Directory"})':
                created_outputs.append(listing.entryname)

    for o in outputs:
        prefix = None
        position = None

        if o.outputBinding.glob.startswith("$(inputs."):
            name = o.outputBinding.glob.split(".")[1]

            if name.endswith(")"):
                name = name[:-1]

            matched_input = next(i for i in inputs if i.id == name)
            inputs.remove(matched_input)

            if isinstance(matched_input.default, dict):
                path = client.workflow_path / Path(matched_input.default["path"])
            else:
                path = Path(matched_input.default)

            path = Path(os.path.abspath(client.path / path)).relative_to(client.path)

            if matched_input.inputBinding:
                prefix = matched_input.inputBinding.prefix
                position = matched_input.inputBinding.position

                if prefix and matched_input.inputBinding.separate:
                    prefix += " "
        else:
            path = Path(o.outputBinding.glob)

        create_folder = False

        check_path = path
        if not (client.path / path).is_dir():
            check_path = path.parent

        if check_path != "." and str(check_path) in created_outputs:
            create_folder = True

        run.outputs.append(
            CommandOutput(
                id=CommandOutput.generate_id(base_id, position),
                position=position,
                prefix=prefix,
                produces=_entity_from_path(client, path, commit),
                create_folder=create_folder,
            )
        )

    for i in inputs:
        prefix = None
        position = None

        if i.inputBinding:
            prefix = i.inputBinding.prefix
            position = i.inputBinding.position

            if prefix and i.inputBinding.separate:
                prefix += " "

        if isinstance(i.default, dict) and "class" in i.default and i.default["class"] in ["File", "Directory"]:
            path = client.workflow_path / Path(i.default["path"])
            path = Path(os.path.abspath(path)).relative_to(client.path)

            run.inputs.append(
                CommandInput(
                    id=CommandInput.generate_id(base_id, position),
                    position=position,
                    prefix=prefix,
                    consumes=_entity_from_path(client, path, commit),
                )
            )
        else:
            run.arguments.append(
                CommandArgument(
                    id=CommandArgument.generate_id(base_id, position),
                    position=position,
                    prefix=prefix,
                    value=str(i.default),
                )
            )

    for a in cmd_line_tool.arguments:
        id_ = CommandArgument.generate_id(base_id, a["position"])
        run.arguments.append(CommandArgument(id=id_, position=a["position"], value=a["valueFrom"]))

    if not persist:
        return run, None

    step_name = "{0}_{1}.yaml".format(uuid.uuid4().hex, secure_filename("_".join(cmd_line_tool.baseCommand)),)

    absolute_path = client.workflow_path / step_name
    path = absolute_path.relative_to(client.path)

    run.path = path
    process_run = ProcessRun.from_run(run, client, path, commit=commit)
    process_run.invalidated = _invalidations_from_commit(client, commit)

    # HACK: This fixes broken SoftwareAgent due to rebases done by users
    if isinstance(process_run.association.agent, Person) or not process_run.association.agent.label.startswith(
        "renku "
    ):
        process_run.association.agent = default_missing_software_agent
    process_run.to_yaml(path=absolute_path)
    client.add_to_activity_index(process_run)
    return process_run, absolute_path


def _migrate_composite_step(client, workflow, path, commit=None):
    """Migrate a composite workflow."""
    if not commit:
        commit = client.find_previous_commit(path)
    run = Run(client=client, path=path, commit=commit)
    rel_path = Path(path).relative_to(client.path)
    label = f"{rel_path}@{commit.hexsha}"
    identifier = sha1(label.encode("utf-8")).hexdigest()
    run._id = Run.generate_id(client, identifier=identifier)

    name = "{0}_migrated.yaml".format(uuid.uuid4().hex)

    run.path = (client.workflow_path / name).relative_to(client.path)

    for step in workflow.steps:
        if isinstance(step.run, dict):
            continue
        else:
            path = client.workflow_path / step.run
            subrun = parse_cwl_cached(str(path))

        subprocess, _ = _migrate_single_step(client, subrun, path, parent_commit=commit)
        run.add_subprocess(subprocess)

    wf = WorkflowRun.from_run(run, client, run.path, commit=commit)

    # HACK: This fixes broken SoftwareAgent due to rebases done by users
    if isinstance(wf.association.agent, Person) or not wf.association.agent.label.startswith("renku "):
        wf.association.agent = default_missing_software_agent
    for p in wf._processes:
        if isinstance(p.association.agent, Person) or not p.association.agent.label.startswith("renku "):
            p.association.agent = default_missing_software_agent
    wf.to_yaml(run.path)
    client.add_to_activity_index(wf)

    return wf, run.path


def _entity_from_path(client, path, commit):
    """Gets the entity associated with a path."""
    client, commit, path = client.resolve_in_submodules(
        client.find_previous_commit(path, revision=commit.hexsha), path,
    )

    entity_cls = Entity
    if (client.path / path).is_dir():
        entity_cls = Collection

    if str(path).startswith(os.path.join(client.renku_home, client.DATASETS)):
        return Dataset.from_yaml(path=client.path / path, client=client, commit=commit)
    else:
        return entity_cls(commit=commit, client=client, path=str(path),)


def _invalidations_from_commit(client, commit):
    """Gets invalidated files from a commit."""
    results = []
    collections = dict()
    for file_ in commit.diff(commit.parents or NULL_TREE):
        # only process deleted files (note they appear as ADDED)
        # in this backwards diff
        if file_.change_type != "A":
            continue
        path_ = Path(git_unicode_unescape(file_.a_path))
        entity = _get_activity_entity(client, commit, path_, collections, deleted=True)

        results.append(entity)

    return results


def _get_activity_entity(client, commit, path, collections, deleted=False):
    """Gets the entity associated with this Activity and path."""
    client, commit, path = client.resolve_in_submodules(commit, path,)
    output_path = client.path / path
    parents = list(output_path.relative_to(client.path).parents)

    collection = None
    members = []
    for parent in reversed(parents[:-1]):
        if str(parent) in collections:
            collection = collections[str(parent)]
        else:
            collection = Collection(client=client, commit=commit, path=str(parent), members=[], parent=collection,)
            members.append(collection)
            collections[str(parent)] = collection

        members = collection.members

    entity_cls = Entity
    if (client.path / path).is_dir():
        entity_cls = Collection

    if str(path).startswith(os.path.join(client.renku_home, client.DATASETS)) and not deleted:
        entity = Dataset.from_yaml(path=client.path / path, client=client, commit=commit)
    else:
        entity = entity_cls(commit=commit, client=client, path=str(path), parent=collection,)

    if collection:
        collection.members.append(entity)

    return entity


_cwl_cache = {}


def parse_cwl_cached(path):
    """Parse cwl and remember the result for future execution."""
    if path in _cwl_cache:
        return _cwl_cache[path]

    cwl = parse_cwl(path)

    _cwl_cache[path] = cwl

    return cwl


def _find_only_cwl_commit(client, path, revision="HEAD"):
    """Find the most recent isolated commit of a cwl file.

    Commits with multiple cwl files are disregarded
    """
    file_commits = list(client.repo.iter_commits(revision, paths=path, full_history=True))

    for commit in file_commits:
        cwl_files = [f for f in commit.stats.files.keys() if f.startswith(client.cwl_prefix) and path.endswith(".cwl")]
        if len(cwl_files) == 1:
            return commit

    raise ValueError("Couldn't find a previous commit for path {}".format(path))



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataverse metadata templates."""
DATASET_METADATA_TEMPLATE = """
{
    "datasetVersion": {
        "metadataBlocks": {
            "citation": {
                "fields": [
                    {
                        "value": "${name}",
                        "typeClass": "primitive",
                        "multiple": false,
                        "typeName": "title"
                    },
                    {
                        "value": ${authors},
                        "typeClass": "compound",
                        "multiple": true,
                        "typeName": "author"
                    },
                    {
                        "value": ${contacts},
                        "typeClass": "compound",
                        "multiple": true,
                        "typeName": "datasetContact"
                    },
                    {
                        "value": [
                            {
                                "dsDescriptionValue": {
                                    "value": "${description}",
                                    "multiple": false,
                                    "typeClass": "primitive",
                                    "typeName": "dsDescriptionValue"
                                }
                            }
                        ],
                        "typeClass": "compound",
                        "multiple": true,
                        "typeName": "dsDescription"
                    },
                    {
                        "value": [],
                        "typeClass": "controlledVocabulary",
                        "multiple": true,
                        "typeName": "subject"
                    }
                ],
                "displayName": "Citation Metadata"
            }
        }
    }
}"""

AUTHOR_METADATA_TEMPLATE = """
{
    "authorName": {
        "value": "${name}",
        "typeClass": "primitive",
        "multiple": false,
        "typeName": "authorName"
    },
    "authorAffiliation": {
        "value": "${affiliation}",
        "typeClass": "primitive",
        "multiple": false,
        "typeName": "authorAffiliation"
    }
}
"""

CONTACT_METADATA_TEMPLATE = """
{
    "datasetContactEmail": {
        "typeClass": "primitive",
        "multiple": false,
        "typeName": "datasetContactEmail",
        "value": "${email}"
    },
    "datasetContactName": {
        "typeClass": "primitive",
        "multiple": false,
        "typeName": "datasetContactName",
        "value": "${name}"
    }
}
"""
